# coding: utf-8


import codecs
import json


class ConfigBase(object):

    def __getitem__(self, key):
        raise NotImplementedError

    def add(self, item):
        raise NotImplementedError

    def extend(self, iterable):
        raise NotImplementedError

    def remove(self, item, recurse):
        raise NotImplementedError

    def get(self, param):
        raise NotImplementedError

    @property
    def dict(self):
        raise NotImplementedError


class CompositeItem(ConfigBase):
    def __init__(self, name, value=None, type_=None, uniq_childs=False):
        assert isinstance(name, basestring), '`name` parameter must be string.'

        assert isinstance(uniq_childs, bool), '`uniq_childs` must be bool.'

        self._items = list()
        self.uniq_childs = uniq_childs

        self.name = name
        self.value = value
        self.type_ = type_

    def __getattr__(self, key):
        items_dict = dict(self._items)
        if key in items_dict:
            return items_dict[key]

    def add(self, item):
        u"""Add nested item."""
        assert isinstance(item, self.__class__), type(item)
        if self.uniq_childs and item.name in dict(self._items):
            raise ValueError(
                u'`{0}` already in `{1}`'.format(item.name, self.name)
            )
        self._items.append((item.name, item))
        return self

    def extend(self, iterable):
        u"""Extend self items with iterable."""
        for item in iterable:
            self.add(item)
        return self

    def remove(self, item, recurse=False):
        u"""Remove nested item."""
        assert isinstance(item, self.__class__)
        lookup = (item.name, item)
        if lookup in self._items:
            self._items.remove(lookup)
        raise ValueError(
            u'`{0}` not in `{1}`'.format(item.name, self.name)
        )

    def get(self, param_name):
        u"""Recursive item get."""
        item_list = list()
        for item_name, item in self._items:
            if item_name == param_name:
                item_list.append(item)
            item_list.extend(
                item.get(param_name)
            )
        return item_list

    @property
    def dict(self):
        u"""Item as dict."""
        result = dict()
        for name, item in self._items:
            if item.value is not None:
                result[name] = item.value
            else:
                result[name] = item.dict
        return result

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return u'<{0}: {1} ({2})>'.format(
            self.name, self.value, self.dict
        )


item = CompositeItem


class Config(ConfigBase):
    def __init__(self, filename, uniq_childs=True):
        assert isinstance(filename, basestring), '`filename` must be string.'
        self.filename = filename
        self.root = CompositeItem('config', uniq_childs=uniq_childs)
        self.parser = json

    def __getattr__(self, name):
        return getattr(self.root, name)

    def extend(self, iterable):
        u"""Extend config with iterable."""
        return self.root.extend(iterable)

    def add(self, item):
        u"""Add item."""
        return self.root.add(item)

    def get(self, param):
        u"""Recursive item get."""
        return self.root.get(param)

    @property
    def dict(self):
        u"""Whole config as dict."""
        return self.root.dict

    @property
    def output(self):
        u"""String to write to the file."""
        return self.parser.dumps(self.dict)

    def _parse_source(self, source):
        u"""Parse source recursive."""
        assert isinstance(source, dict)

        node_list = list()

        for name, value in source.iteritems():
            item = CompositeItem(name)
            if not isinstance(value, dict):
                item.value = value
            else:
                subitems = self._parse_source(value)
                item.extend(subitems)
            node_list.append(item)
        return node_list

    def load(self):
        u"""Read config from file."""
        with codecs.open(self.filename, 'r', encoding='utf-8') as fd:
            source = self.parser.loads(fd.read())
            self.root.extend(self._parse_source(source))

    def save(self):
        u"""Write config to the file."""
        with codecs.open(self.filename, 'w', encoding='utf-8') as fd:
            fd.write(self.output)
